/* 
   cadaver, command-line DAV client
   Copyright (C) 1999, Joe Orton <joe@orton.demon.co.uk>
                                                                     
   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 of the License, or
   (at your option) any later version.
  
   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.
  
   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
   
   $Id: cmdline.c,v 1.1 1999/12/19 01:26:12 joe Exp $
*/

#include <config.h>

#ifdef HAVAE_UNISTD_H
#include <unistd.h>
#endif
#ifdef HAVE_STDLIB_H
#include <stdlib.h>
#endif
#include <stdio.h>

#ifdef HAVE_STRING_H
#include <string.h>
#endif

#include "httpdav.h"
#include "basename.h"
#include "fnmatch.h"

#include "cadaver.h"
#include "cmdline.h"

static int has_glob_pattern( const char *str ) {
    const char *pnt;
    for( pnt = str; *pnt != '\0'; pnt++ )
	if( *pnt=='*' || *pnt=='[' || *pnt=='?' )
	    return 1;
    return 0;
}

/* Gets the next token for parse_command...
 * Starts at position in line.
 * DFA states:
 *   0: chewing leading whitespace
 *   1: chewing characters, normally
 *   2: chewing characters in a quote
 *   3: just got a backslash in normal chew
 *   4: just got a backslash in a quoted chew
 *   9: acceptance state
 * 
 * State diagram is left as an exercise to the readers, since 
 * I'm not going to draw it in ASCII. ;)
 *
 * Returns the token, or NULL on end-of-line.
 * Position updated to point after token.
 */
static char *gettoken( const char *line, const char **pointer ) {
    const char *pnt;
    int state = 0, pos = 0;
    char buf[BUFSIZ], quote;
    
    pnt = *pointer;
    
#define ISQUOTE(x) (x=='\'' || x=='\"')
#define ISWHITE(x) (x==' ' || x=='\t')

     while( *pnt != '\0' && state != 9 && pos < BUFSIZ) {
	 switch( state ) {
	 case 0: /* leading whitespace chew */
	     if( ISQUOTE(*pnt) ) {
		 state = 2;
		 quote = *pnt;
	     } else if( !ISWHITE(*pnt) ) {
		 buf[pos++] = *pnt;
		 state = 1;
	     }
	     break;
	 case 1: /* normal chew */
	     if( ISWHITE(*pnt) ) {
		 state = 9;
	     } else if( *pnt == '\\' ) {
		 state = 3;
	     } else {
		 buf[pos++] = *pnt;
	     }
	     break;
	 case 2: /* quoted chew */
	     if( *pnt == quote ) {
		 state = 9;
	     } else if( *pnt == '\\' ) {
                state = 4;
	     } else {
		 buf[pos++] = *pnt;
	     }
	     break;
	 case 3: /* chew an escaped literal */
	     buf[pos++] = *pnt;
	     state = 1;
	     break;
	 case 4:
	     buf[pos++] = *pnt;
	     state = 2;
	     break;
	 }
	 pnt++;
     }
     
#undef ISQUOTE
#undef ISWHITE

     /* overflowed the buffer */
     if( pos == BUFSIZ ) return NULL;
     
     buf[pos] = '\0';
     *pointer = pnt;
     if( pos > 0 ) {
#if 0
	 /* a little hack; does env. var expansion...
	  * 1) is this really useful?
	  * 2) this should be done in parse_command not gettoken
	  */
	 if( buf[0] == '$' ) {
	     char *val = getenv(&buf[1]);
	     if( val ) return strdup(val);
	 }
#endif
	 return strdup(buf);
     } else {
	 return NULL;
     }
}


char **parse_command( const char *line, int *count ) {
    struct proto_file *list, *current, *next; /* for storing the file list if needed */
    int got_list = 0; /* -1: failed, 0: not done yet, 1: success */
    char *token, **tokens = NULL;
    const char *pnt = line;
    int numtokens = 0, matches = 0;

#define ADDTOK(x) 						\
{								\
    tokens = realloc( tokens, ++numtokens*sizeof(char *) );	\
    tokens[numtokens-1] = x;					\
}

    for(;;) {
	token = gettoken( line, &pnt );
	if( token ) { 
	    if( has_glob_pattern( token ) && have_connection ) {
		fflush( stdout );
		if( got_list == 0 ) {
		    printf( "[Matching... " );
		    if( dav_fetch( path, 1, &list ) == PROTO_OK ) {
			printf( "got resource list]\n" );
			got_list = 1;
		    } else {
			printf( "failed to fetch resource list]\n" );
			got_list = -1;
		    }
		}
		matches = 0;
		if( got_list == 1 ) {
		    for( current = list; current!=NULL; current=current->next ) {
			char *bname = base_name( current->filename );
			if( fnmatch( token, bname, 0 ) == 0 ) {
			    matches++;
			    ADDTOK( strdup(bname) );
			}
		    }
		}
		if( matches > 0 ) {
		    free( token );
		} else {
		    ADDTOK( token );
		}
	    } else {
		ADDTOK( token );
	    }
	} else {
	    break;
	}
    }

    if( got_list == 1 ) {
	for( current = list; current != NULL; current = next ) {
	    next = current->next;
	    free( current );
	}
    }

    *count = numtokens;
    /* add a NULL at the end of the list */
    ADDTOK(NULL);

    return tokens;
}
